package com.wechi.adweb.bridge.google.gtm.service;

import static com.wechi.adweb.bridge.google.common.ResourceDTO.parseId;

import com.google.api.client.googleapis.javanet.GoogleNetHttpTransport;
import com.google.api.client.json.gson.GsonFactory;
import com.google.api.client.util.Lists;
import com.google.api.services.tagmanager.TagManager;
import com.google.api.services.tagmanager.TagManagerScopes;
import com.google.api.services.tagmanager.model.*;
import com.google.auth.http.HttpCredentialsAdapter;
import com.google.auth.oauth2.GoogleCredentials;
import com.google.common.annotations.VisibleForTesting;
import com.wechi.adweb.bridge.exception.DataException;
import com.wechi.adweb.bridge.google.gtm.dto.GTMAccountDTO;
import com.wechi.adweb.bridge.google.gtm.dto.GTMContainerDTO;
import com.wechi.adweb.bridge.google.gtm.dto.GTMWorkspaceDTO;
import com.wechi.adweb.bridge.util.JsonUtils;

import lombok.extern.slf4j.Slf4j;

import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.security.GeneralSecurityException;
import java.util.Collections;
import java.util.List;

import javax.annotation.PostConstruct;

/**
 * @author wfansh
 */
@Slf4j
@Service
public class GTMService {

    @Value("${google.service.account.key}")
    private String serviceAccountKey;

    @Value("${google.gtm.container.admins}")
    private List<String> containerAdmins;

    private TagManager tagManager;

    @PostConstruct
    private void init() throws GeneralSecurityException, IOException {
        // Service account authorization;
        GoogleCredentials credentials =
                GoogleCredentials.fromStream(
                                this.getClass()
                                        .getClassLoader()
                                        .getResourceAsStream(serviceAccountKey))
                        .createScoped(TagManagerScopes.all());

        this.tagManager =
                new TagManager.Builder(
                                GoogleNetHttpTransport.newTrustedTransport(),
                                GsonFactory.getDefaultInstance(),
                                new HttpCredentialsAdapter(credentials))
                        .build();
    }

    public List<GTMAccountDTO> listGTMAccounts() throws DataException {
        try {
            List<GTMAccountDTO> gtmAccounts = Lists.newArrayList();
            for (Account account : tagManager.accounts().list().execute().getAccount()) {
                GTMAccountDTO gtmAccount =
                        GTMAccountDTO.builder()
                                .id(account.getAccountId())
                                .resourceName(account.getPath())
                                .displayName(account.getName())
                                .supportUserPermissions(
                                        account.getFeatures().getSupportUserPermissions())
                                .supportMultipleContainers(
                                        account.getFeatures().getSupportMultipleContainers())
                                .build();

                // Adds into the list;
                gtmAccounts.add(gtmAccount);
            }

            log.info("listGTMAccounts : {}", JsonUtils.toJson(gtmAccounts));
            return gtmAccounts;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    public List<GTMContainerDTO> listGTMContainers(String accountResourceName)
            throws DataException {
        try {
            List<GTMContainerDTO> gtmContainers = Lists.newArrayList();
            for (Container container :
                    tagManager
                            .accounts()
                            .containers()
                            .list(accountResourceName)
                            .execute()
                            .getContainer()) {
                GTMContainerDTO gtmContainer =
                        GTMContainerDTO.builder()
                                .id(container.getContainerId())
                                .resourceName(container.getPath())
                                .displayName(container.getName())
                                .accountId(container.getAccountId())
                                .domainNames(container.getDomainName())
                                .publicId(container.getPublicId())
                                .tagIds(container.getTagIds())
                                .notes(container.getNotes())
                                .usageContext(container.getUsageContext())
                                .fingerprint(container.getFingerprint())
                                .build();

                // Adds into the list;
                gtmContainers.add(gtmContainer);
            }

            log.info(
                    "listGTMContainers for {} : {}",
                    accountResourceName,
                    JsonUtils.toJson(gtmContainers));
            return gtmContainers;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    public GTMContainerDTO createGTMContainer(String accountResourceName, String displayName)
            throws DataException {
        try {
            Container container = new Container();
            container.setName(displayName);
            container.setUsageContext(Collections.singletonList("web"));

            ContainerFeatures containerFeatures = new ContainerFeatures();
            containerFeatures.setSupportUserPermissions(false);
            container.setFeatures(containerFeatures);

            Container createdContainer =
                    tagManager
                            .accounts()
                            .containers()
                            .create(GTMAccountDTO.toResourceName(accountResourceName), container)
                            .execute();
            log.info("Created container {}", JsonUtils.toJson(createdContainer));

            // Converts and returns.
            return GTMContainerDTO.builder()
                    .id(createdContainer.getContainerId())
                    .resourceName(createdContainer.getPath())
                    .displayName(createdContainer.getName())
                    .accountId(parseId(accountResourceName))
                    .domainNames(createdContainer.getDomainName())
                    .publicId(createdContainer.getPublicId())
                    .tagIds(createdContainer.getTagIds())
                    .notes(createdContainer.getNotes())
                    .usageContext(createdContainer.getUsageContext())
                    .fingerprint(container.getFingerprint())
                    .build();
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    /**
     * See
     * https://developers.google.com/tag-platform/tag-manager/api/v2/reference/accounts/user_permissions
     */
    @VisibleForTesting
    void setupContainerAccess(String accountResourceName, String containerId) throws DataException {
        try {
            // 1. Container access.
            ContainerAccess containerAccess = new ContainerAccess();
            containerAccess.setContainerId(containerId);
            containerAccess.setPermission("publish");

            // 2. Account access - required in user permission request.
            AccountAccess accountAccess = new AccountAccess();
            accountAccess.setPermission("admin");

            // 3. Combines as user permission.
            UserPermission userPermission = new UserPermission();
            userPermission.setContainerAccess(Collections.singletonList(containerAccess));
            userPermission.setAccountAccess(accountAccess);

            for (String containerAdmin : containerAdmins) {
                // One email address per request.
                userPermission.setEmailAddress(containerAdmin);
                UserPermission createdUserPermission =
                        tagManager
                                .accounts()
                                .userPermissions()
                                .create(accountResourceName, userPermission)
                                .execute();
                log.info("Created user permission {}", JsonUtils.toJson(createdUserPermission));
            }
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    @VisibleForTesting
    GTMWorkspaceDTO getDefaultWorkspace(String containerResourceName) throws DataException {
        try {
            Workspace workspace =
                    tagManager
                            .accounts()
                            .containers()
                            .workspaces()
                            .list(containerResourceName)
                            .execute()
                            .getWorkspace()
                            .stream()
                            .filter(w -> w.getName().equalsIgnoreCase("Default Workspace"))
                            .findFirst()
                            .get();

            log.info(
                    "Get default workspace for {} : {}",
                    containerResourceName,
                    JsonUtils.toJson(workspace));

            // Converts and returns.
            return GTMWorkspaceDTO.builder()
                    .id(workspace.getWorkspaceId())
                    .resourceName(workspace.getPath())
                    .displayName(workspace.getName())
                    .accountId(workspace.getAccountId())
                    .containerId(workspace.getContainerId())
                    .description(workspace.getDescription())
                    .fingerprint(workspace.getFingerprint())
                    .build();
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }
}
